
### Copyright (c) 2018 Roman T Oliynyk, under MIT licence. Refer to https://opensource.org/licenses/MIT ###

# this R script takes simulation files: CohortHeritabilityAnalysis.csv CohortStatisticalPower.csv EnvironmentArray.csv HeritabilityAnalysis.csv ScatterData.csv StatisticalPower.csv 
# and produces the Statistical Power files; Files starting with 'S' are smoothed using LOESS.
# resulting files: CasesNeededStatisticalPower.csv CasesNeededCumulativeStatisticalPower.csv CasesNeededCohortStatisticalPower.csv SCasesNeededCohortStatisticalPower.csv SCasesNeededStatisticalPower.csv SCasesNeededCumulativeStatisticalPower.csv SRelativeMafCohortStatisticalPower.csv SRelativeMafStatisticalPower.csv

bSavePDFs = FALSE  #if you also want to have R produce a preview PDFs, change to TRUE

if(  bSavePDFs ) {
  if (!require("ggplot2")) {
    install.packages("ggplot2", dependencies = TRUE)
    library(ggplot2)
  }
  if (!require("RColorBrewer")) {
    install.packages("RColorBrewer", dependencies = TRUE)
    library(RColorBrewer)
  }
}

OneBil = 1000000000.0


CountNCPCore <- function(mafCases, mafControls) {
  mafAverage = ( mafCases + mafControls ) / 2
  beta = log((mafCases / (1 - mafCases)) / (mafControls / (1 - mafControls)))
  NCPCore = 2 * (1.0 - mafAverage) * mafAverage * 0.25 * (beta^2)
  return (NCPCore)
}

CalculatePower <- function(N, NCPCore) {
  NCP = N * NCPCore
  PSign = 0.99999995
  
  StatPower = 1 - pf(qf(PSign, 1, N - 4), 1, N - 4, NCP)
  
  return( StatPower )
}


EstimateCaseNumber  <- function( mafCases, mafControls ) {
  N = OneBil # start wit 1Billion cases/controls needed

  StatPower = 1
  
  if( mafCases - mafControls < 0 ) {
    return (OneBil)  # cases have smaller OR than controls, noise of course, but can't 
  }
    
  NCPCore = CountNCPCore( mafCases, mafControls)
 
  while(StatPower > 0.8) {
    StatPower = CalculatePower( N, NCPCore )
    if( StatPower > 0.8 ) {
      N = N * 0.5
    }
  }
  
  while(StatPower < 0.8 && N < OneBil ) {
    N = N * 1.01
    StatPower = CalculatePower( N, NCPCore )
  }
  while(StatPower > 0.8 && N < OneBil ) {
    N = N * 0.999
    StatPower = CalculatePower( N, NCPCore )
  }
  while(StatPower < 0.8 && N < OneBil ) {
    N = N * 1.0001
    StatPower = CalculatePower( N, NCPCore )
  }
  return( N )  # we use equal number of cases and controls throuthout this model, half cases / half controls
}


SaveDataFrameToFile <- function(DataFr, pathname, fileName) {
  # save all incidence, original and with 5, 10,   
  destfile <- paste(pathname, "/", fileName, sep="")
  write.table(format( DataFr, digits=4, scientific=TRUE, trim = TRUE),
              file = destfile,
              row.names = FALSE,
              col.names = TRUE,
              sep = ",",
              quote = FALSE )
}


SaveDataSetToFile <- function( age, df, pathname, filename, nAlleles) {
  if( nAlleles == 1) {
    IncDataFr = data.frame( age, df[1])
  } else {
    IncDataFr = data.frame( age, df[1],df[2],df[3],df[4],df[5],df[6],df[7],df[8],df[9])
  }
  SaveDataFrameToFile(IncDataFr, pathname, filename) 
  rm( IncDataFr);
}



main <- function() {
  args <- commandArgs(trailingOnly = TRUE)
  pathname <- args[1]
  title <- args[2]
  sourceCsv <- args[3]
  nLimitAgeYoung <- as.numeric(args[4])
  nLimitAgeOld <- as.numeric(args[5])
  nCohortAgeForOlderControls <- as.numeric(args[6])
  
  
  sourcefile <- paste(pathname,"/",sourceCsv, sep="")
  statData = read.csv( sourcefile )
  
  sourcefile <- paste(pathname,"/../../","StatisticalPowerAlleleParameters.txt", sep="")
  statConfig = read.csv( sourcefile )
  
  szLimitText = "Cohort"  # empty Limit text as a default
  
  nLimitAgeFrontCorrect = 1
  limitAgeCorrect = length(statData$Age)
  
  if( !is.na(nLimitAgeOld) ) {  # limit starting age
    for(i in 1:(length(statData$Age)) ) {
      if(statData$Age[i] == nLimitAgeOld) {
        limitAgeCorrect = i;
        break;
      }
    }     
  } 
  
  if( !is.na(nLimitAgeYoung) ) {  # limit top age
    for(i in 1:limitAgeCorrect ) {
      if(statData$Age[i] == nLimitAgeYoung) {
        nLimitAgeFrontCorrect = i;
        break;
      }
    }     
  } 
  
  statData <- statData[nLimitAgeFrontCorrect:limitAgeCorrect,]
  
  bDoCumulative = FALSE
  #if( "StatisticalPower.csv" == sourceCsv) {
  #  bDoCumulative = TRUE
  #}
  
  age <- statData$Age
  
  nTrackedAlleles <- length(statConfig$Beta)
  dbSmoothingSpan = 0.5

  if( !is.na(nCohortAgeForOlderControls) && "CohortStatisticalPower.csv" == sourceCsv ) { #if we need the cohort and instantaneous comparison Younger Cases / Older Controls
    
    nAgeOffset = (nCohortAgeForOlderControls - age[1])
    if( nAgeOffset < 0 ) {
      nAgeOffset = 0
    }
    
    dfCasesSameAge <- data.frame(matrix(ncol = nTrackedAlleles, nrow = length(age) - nAgeOffset))
    dfCasesYoungOld <- data.frame(matrix(ncol = nTrackedAlleles, nrow = length(age) - nAgeOffset))
    dfMultCasesYoungOld <- data.frame(matrix(ncol = nTrackedAlleles, nrow = length(age) - nAgeOffset))
    dfMultCasesSameAge <- data.frame(matrix(ncol = nTrackedAlleles, nrow = length(age) - nAgeOffset))
    ageChopped <- age[(nAgeOffset+1):length(age)]
    
    for(i in 1:(nTrackedAlleles) ) {
      mafControls <- statData[,(i-1)*5 + 3] * statConfig$Beta[i]
      if( nTrackedAlleles > 1 ) {  # if we have only one entry, the we have one allele analyzed in as many copies as possible for high precision analysis, don't smooth if 1 Allele
        smooth=loess(mafControls ~ age, span=dbSmoothingSpan)
        mafControls <- smooth$fitted
      }    
      mafCases <- statData[,(i-1)*5 + 4] * statConfig$Beta[i]
      if( nTrackedAlleles > 1 ) {  # if we have only one entry, the we have one allele analyzed in as many copies as possible for high precision analysis, don't smooth if 1 Allele
        smooth=loess(mafCases ~ age, span=dbSmoothingSpan)
        mafCases <- smooth$fitted
      }
      
      mafAtAge <- 0
      
      nCasesNeeded <- rep(NA, length(age))
      nCasesNeededSameAge <- rep(NA, length(age))
      nCasesMult <- rep(NA, length(age))
      nCasesSameAgeMult <- rep(NA, length(age))
      
      nBaselineCasesNeeded = 0
      
      
      for(ageelement in 1:(length(age)) ) {
        if(  age[ageelement] >= nCohortAgeForOlderControls ) {
          if(  age[ageelement] == nCohortAgeForOlderControls ) {
            mafAtAge <- mafCases[ageelement]
          }
          
          nNum = EstimateCaseNumber(mafAtAge, mafControls[ageelement])
          nNumSameAge = EstimateCaseNumber( mafCases[ageelement], mafControls[ageelement])
          
          if( nNumSameAge < (OneBil - 100000000.0)) {
            nCasesNeededSameAge[ageelement]  <- nNumSameAge
          } 
          
          if( nNum < (OneBil - 100000000.0)) {
            nCasesNeeded[ageelement]  <- nNum
            
            if(  age[ageelement] == nCohortAgeForOlderControls ) {
              nBaselineCasesNeeded = nNum
            }
            if( nBaselineCasesNeeded > 1 ) {
              nCasesMult[ageelement] = nNum / nBaselineCasesNeeded
              nCasesSameAgeMult[ageelement] = nNumSameAge / nBaselineCasesNeeded
            }
          } 
        }
      }
      dfCasesSameAge[i] <- nCasesNeededSameAge[(nAgeOffset+1):length(age)]
      dfCasesYoungOld[i] <- nCasesNeeded[(nAgeOffset+1):length(age)]
      dfMultCasesYoungOld[i] <- nCasesMult[(nAgeOffset+1):length(age)]
      dfMultCasesSameAge[i] <- nCasesSameAgeMult[(nAgeOffset+1):length(age)]
    }
    
    destfile <- paste( "CasesNeededYoungOld",sourceCsv, sep="")
    print( destfile) 
    SaveDataSetToFile( ageChopped, dfCasesYoungOld, pathname, destfile, nTrackedAlleles);
    
    destfile <- paste( "CasesMultipleYoungOld",sourceCsv, sep="")
    SaveDataSetToFile( ageChopped, dfMultCasesYoungOld, pathname, destfile, nTrackedAlleles);
    
    destfile <- paste( "CasesMultipleSameAge",sourceCsv, sep="")
    SaveDataSetToFile( ageChopped, dfMultCasesSameAge, pathname, destfile, nTrackedAlleles);

    destfile <- paste( "CasesNeeded",sourceCsv, sep="")
    SaveDataSetToFile( ageChopped, dfCasesSameAge, pathname, destfile, nTrackedAlleles);
    
  }
  
  # return(0)

}

##### !!!!!!!!!!!!!!!!!!! ***
main() 
##### !!!!!!!!!!!!!!!!!!! ***